// CHPegBoard.m -- Model object for PegSolitaire.  Contains peg count, layout,
//                 and adjacency information
//
// Copyright (c) 2004 CocoaHeads.  All rights reserved.
// Released under the BSD license

#import "CHPegBoard.h"
#import "CHPegBoardView.h"

@implementation CHPegBoardView


// designated initializer for our superclass.
// set up the pegStates (on/off) array here, since this is still
// 'scaffolding to get stuff working' mode

- (id) initWithFrame: (NSRect) frame
{
    if (self = [super initWithFrame: frame]) {
    }

    return (self);

} // initWIthFrame


// clean up our mess

- (void) dealloc
{
    int i;
    for (i = 0; i < [pegBoard spaceCount]; i++) {
        [pegPaths[i] release];
    }

    [pegBoard release];

    [super dealloc];

} // dealloc



// nib file has been loaded and all connections made in interface builder
// have been made

- (void) awakeFromNib
{
    // we want mouse motion events so we can highlight pegs as the user
    // mouses over them

    [[self window] setAcceptsMouseMovedEvents: YES];

} // awakeFromNib



// utility method that ideally should be in some common library
// or provided by Apple.
// Given two rectangles, center one inside of the other

- (NSRect) centerRect: (NSRect) smallRect
               inRect: (NSRect) bigRect
{
    NSRect centerRect;
    centerRect.size = smallRect.size;

    centerRect.origin.x = (bigRect.size.width - smallRect.size.width) / 2.0;
    centerRect.origin.y = (bigRect.size.height - smallRect.size.height) / 2.0;

    return (centerRect);

} // centerRect



// create a maximal-sized square that fits inside of our view.

- (NSRect) makeCenterSquareRect
{
    // our bounding rectangle
    NSRect bounds = [self bounds];

    // figure out whether the width or height is smaller.  That's our
    // controlling dimension
    float minDimension;
    minDimension = MIN (bounds.size.width, bounds.size.height);

    // make a square the proper size.  This has an origin of zero, but
    // can be based anywhere since [self centerRect] will change the 
    // x and y
    NSRect centerSquare = NSMakeRect (0.0, 0.0, minDimension, minDimension);

    // and now center it in our bounds
    centerSquare = [self centerRect: centerSquare
                         inRect: bounds];

    // and skootch it in to make it not coincident with the view border.
    // makes it look a little nicer
    centerSquare = NSInsetRect (centerSquare, 5, 5);

    return (centerSquare);
    
} // makeCenterSquareRect


// something generic is just screaming to escape from this.
//
// Rather than calculating each peg's location directly, it 
// spreads each peg on the row out, based on the left, with the Y value
// set correctly.  For triangular peg solitaire, it would look like:
// +--+
// |  |
// +--+--+
// |  |  |
// +--+--+--+
// |  |  |  |
// +--+--+--+--+
// |  |  |  |  |
// +--+--+--+--+--+
// |  |  |  |  |  |
// +--+--+--+--+--+
// then each row of rectangles is centered in the 'center rect' that
// is our play area.
//
// this technique (build X number of rectangles for each row, then
// center the union in the peg rectangle) would also help in laying out
// the 'line o' pegs' and the 'cross' patterns

- (void) recalcPegPathsForRect: (NSRect) squareRect
{
    // first make all the rects flush-left

    // these magic constants were determined via tweaking and fiddling

    float pegRadius = squareRect.size.width / ([pegBoard rowCount] + 3) ;
    float pad = (squareRect.size.width - (pegRadius * [pegBoard maxRowSize])) / ([pegBoard rowCount] + 3);
    float rowHeight = squareRect.size.height / ([pegBoard rowCount] + 1);
    float x, y;

    // union rectangle for each row
    NSRect pegRects[[pegBoard spaceCount]];
    NSRect unions[50];

    int i = 0; // which peg we're looking at

    float baseY;
    baseY = ((squareRect.size.height) - ([pegBoard rowCount] * pegRadius + ([pegBoard rowCount] - 1) * pad)) / 2;

    int row;
    for (row = 0; row < [pegBoard rowCount]; row++) {
        y = baseY + pad + rowHeight * row;
        x = 0;


        // do the pegs of the row.
        int peg;
        for (peg = 0; peg < [pegBoard pegsForRow: row]; peg++) {

            pegRects[i] = NSMakeRect (x, y, pegRadius, pegRadius);
            x += pegRadius + pad;
            
            if (peg == 0) {
                unions[row] = pegRects[i];
            } else {
                unions[row] = NSUnionRect (unions[row], pegRects[i]);
            }
            
            i++;
        }
    }
    
    // now center each rect in our useful area
    
    NSRect centerSquare = [self makeCenterSquareRect];
    i = 0;
    
    for (row = 0; row < [pegBoard rowCount]; row++) {
        NSRect rect = [self centerRect: unions[row]  inRect: centerSquare];
        
        // now update the x values for the row, and update the y
        // for the start of the center rect
        int peg;
        for (peg = 0; peg < [pegBoard pegsForRow: row]; peg++) {
            pegRects[i].origin.x += rect.origin.x + centerSquare.origin.x;
            pegRects[i].origin.y += centerSquare.origin.y;
            i++;
        }
    }

    // make the bezier path for each peg
    for (i = 0; i < [pegBoard spaceCount]; i++) {
        [pegPaths[i] release];
        pegPaths[i] = [NSBezierPath bezierPathWithOvalInRect: pegRects[i]];
        [pegPaths[i] retain];
    }

} // recalcPegPathsForRect



// given a peg index, draw it.

- (void) drawPeg: (int) pegIndex
        selected: (BOOL) isSelected
{
    assert (pegIndex>= 0 && pegIndex< [pegBoard spaceCount]);

    // right now, just make 'on' pegs black, and 'off' pegs white
    // eventually we can do much snazzier pegs

    if (isSelected) {
        [[NSColor yellowColor] setFill];
        [[NSColor redColor] setStroke];
    } else {
        if (pegStates[pegIndex]) {
            [[NSColor blackColor] setFill];
            [[NSColor whiteColor] setStroke];
        } else {
            [[NSColor whiteColor] setFill];
            [[NSColor blackColor] setStroke];
        }
    }
    
    [pegPaths[pegIndex] fill];
    [pegPaths[pegIndex] stroke];

} // drawPeg



// given a rectangle, draw the pegboard in it

- (void) drawBoardInRect: (NSRect) rect
{
    // draw the base board, using a path of a triangle
    // since not an alloc or a copy, this can be considered to be
    // autoreleased and we don't have to worry about it being cleaned up

#if 0
    NSBezierPath *path = [NSBezierPath bezierPath];

    // move to the bottom-left
    [path moveToPoint: rect.origin];

    // line to the top-middle
    NSPoint point;
    point.x = rect.origin.x + rect.size.width / 2.0;
    point.y = rect.origin.y + rect.size.height;
    [path lineToPoint: point];

    // line to the bottom-right
    point.x = rect.origin.x + rect.size.width;
    point.y = rect.origin.y;
    [path lineToPoint: point];

    // and close the path by adding a line back to the bottom left
    [path closePath];
#endif

    NSBezierPath *path = [pegBoard pathForBoard];

    NSAffineTransform *transform = [NSAffineTransform transform];
    [transform translateXBy: rect.origin.x  yBy: rect.origin.y];
    [transform scaleXBy: rect.size.width  yBy: rect.size.height];
    
    [path transformUsingAffineTransform: transform];

    [[NSColor brownColor] set];
    [path fill];

    [[NSColor blackColor] set];
    [path stroke];

} // drawBoardInRect



// this method is called by Cocoa whenever the view needs to be redrawn

- (void) drawRect: (NSRect) rect
{
    // get the bounds, since the rect might be a smaller area to redraw
    NSRect bounds = [self bounds];

    // make a nice white background to draw on
    [[NSColor whiteColor] set];
    [NSBezierPath fillRect: bounds];

    // this area is where the triangular pegboard will go
    NSRect squareRect = [self makeCenterSquareRect];

    if (recalcPegPaths) {
        [self recalcPegPathsForRect: squareRect];
        recalcPegPaths = NO;
    }

    // draw the board
    [self drawBoardInRect: squareRect];

    // draw the pegs on top of the board
    int i;
    for (i = 0; i < [pegBoard spaceCount]; i++) {
        [self drawPeg: i
            selected: (i == selectedPeg)];
    }

    // draw any dragged peg on top of everything

    if (dragImage != nil) {
        [dragImage dissolveToPoint: dragPoint  fraction: 1.0];
    }

    [[NSColor blackColor] set];
    [NSBezierPath strokeRect: bounds];

} // drawRect



// we accept first responder status.  This is necessary to receive
// mouse motion events

- (BOOL) acceptsFirstResponder
{
    return (YES);

} // acceptsFirstResponder



- (int) pegIndexAtPoint: (NSPoint) point
{
    // see if it is over a peg
    int pegIndex = CHPegBoard_NO_PEG;

    int i;
    for (i = 0; i < [pegBoard spaceCount]; i++) {
        if ([pegPaths[i] containsPoint: point]) {
            pegIndex = i;
            break;
        }
    }

    return (pegIndex);

} // pegIndexAtPoint



// creates an image, with a clear background, of a peg with a drop-shadow
// underneath it.
// The dragLocation is adjusted so that if the image is drawn with its
// bottom-left at that point, the peg will appear to be under the mouse
// where the user clicked.

- (NSImage *) makeDragImageForPeg: (int) pegIndex
                        dragPoint: (NSPoint *) dragLocation
{
    assert (pegIndex != CHPegBoard_NO_PEG);
    assert (pegIndex >= 0 && pegIndex < [pegBoard spaceCount]);

    // the enclosing rectangle for the peg.
    NSRect pathBounds = [pegPaths[pegIndex] bounds];

    // the size of the dragged image is the size of the peg,
    // plus a skootch so that the stroked line around the peg
    // stays round.  Plus some space for the drop shadow
#define IMAGE_EXPANSION 15

    // expand it to give us room for the drop image
    NSRect bounds;
    bounds = NSInsetRect (pathBounds, -IMAGE_EXPANSION, -IMAGE_EXPANSION);

    // actually make the image
    NSImage *image;
    image = [[NSImage alloc] initWithSize: bounds.size];

    // make a drop shadow (this is new in Panther)
    NSShadow *shadow = [[NSShadow alloc] init];
    [shadow setShadowBlurRadius: 15.0];
    [shadow setShadowOffset: NSMakeSize (10, -10)];

    // draw into the image
    [image lockFocus]; {
        // all drawing will get a shadow
        [shadow set];

        // since the peg wants to draw at the point it was defined at,
        // and the image wants us to draw at (0, 0), shift the
        // coordinate system over so the drawing does happen at (0, 0)
        NSAffineTransform *transform = [NSAffineTransform transform];
        [transform translateXBy: -bounds.origin.x
                   yBy: -bounds.origin.y];
        [transform concat];
        
        // make it transparent where drawing doesn't touch
        [[NSColor clearColor] set];
        [NSBezierPath fillRect: bounds];
        
        [self drawPeg: pegIndex  selected: NO];

    } [image unlockFocus];

    // offset the given drag location to account for the extra padding
    // put on the image
    dragLocation->x = pathBounds.origin.x - IMAGE_EXPANSION;
    dragLocation->y = pathBounds.origin.y - IMAGE_EXPANSION;

    [shadow release];
    
    return ([image autorelease]);

} // makeDragImageForPeg



// handle a mouse motion event by highlighting the peg under the mouse

- (void) mouseMoved: (NSEvent *) event
{
    // get the mouse position in view coordinates
    NSPoint mouse;
    mouse = [self convertPoint: [event locationInWindow]  fromView: nil];

    // only redraw if the currently selected peg (if any) has changed
    int candidatePeg = [self pegIndexAtPoint: mouse];

    if (candidatePeg != CHPegBoard_NO_PEG) {
        // only allow occupied spots
        if (!pegStates[candidatePeg]) {
            candidatePeg = CHPegBoard_NO_PEG;
        }
    }

    if (selectedPeg != candidatePeg) {
        selectedPeg = candidatePeg;
        [self setNeedsDisplay: YES];
    }

} // mouseMoved



- (void) mouseDown: (NSEvent *) event
{
    // get the mouse position in view coordinates
    NSPoint mouse;
    mouse = [self convertPoint: [event locationInWindow]  fromView: nil];

    dragPeg = [self pegIndexAtPoint: mouse];

    // if we're over an occupied location, start a drag

    if (dragPeg != CHPegBoard_NO_PEG
        && pegStates[dragPeg]) {

        // make the image of a peg with a drop shadow
        dragImage = [self makeDragImageForPeg: dragPeg
                          dragPoint: &dragPoint];
        [dragImage retain];

        // hang on to the mouse location so the image can be moved
        // by the delta with the next mouse position
        lastMouse = mouse;

        // clear out the selected peg.  This will be the current drop
        // destination
        selectedPeg = CHPegBoard_NO_PEG;

        // clear out the source slot, since the peg that lives there is
        // currently moving
        pegStates[dragPeg] = NO;

        [self setNeedsDisplay: YES];
    }

} // mouseDown


- (void) mouseDragged: (NSEvent *) event
{
    // get the mouse position in view coordinates
    NSPoint mouse;
    mouse = [self convertPoint: [event locationInWindow]  fromView: nil];

    if (dragImage != nil) {

        // update the drag image
        float deltaX, deltaY;
        deltaX = mouse.x - lastMouse.x;
        deltaY = mouse.y - lastMouse.y;

        dragPoint.x += deltaX;
        dragPoint.y += deltaY;
        
        // remember the mouse position so we can calculate the delta
        // next time we get a drag
        lastMouse = mouse;

        // highlight the destination if we're over an empty cell
        selectedPeg = [self pegIndexAtPoint: mouse];

        // make sure the destination isn't already occupied
        if (selectedPeg != CHPegBoard_NO_PEG
            && pegStates[selectedPeg]) {

            selectedPeg = CHPegBoard_NO_PEG;
        }

        [self setNeedsDisplay: YES];
    }

} // mouseDragged



- (void) mouseUp: (NSEvent *) event
{
    if (dragImage != nil) {

        // if we're over a peg, return it from whence it came
        if (selectedPeg == CHPegBoard_NO_PEG) {
            pegStates[dragPeg] = YES;

        } else {
            // otherwise 'move' the peg to the new stop
            pegStates[selectedPeg] = YES;
        }

        // clear the selection
        selectedPeg = CHPegBoard_NO_PEG;

        // get rid of the drag image
        [dragImage release];
        dragImage = nil;

        // and redraw to get rid of the drag indicator
        [self setNeedsDisplay: YES];
    }

} // mouseUp



// this gets called by Cocoa when the view gets resized.
// when that happens, we need to recalculate the paths for the pegs

- (void) setFrame: (NSRect) frame
{
    [super setFrame: frame];

    recalcPegPaths = YES;

} // setFrame


- (void) setPegBoard: (CHPegBoard *) newPegBoard
{
    [newPegBoard retain];
    [pegBoard release];

    pegBoard = newPegBoard;

    int i;
    // alternate them on and off, for initial development
    for (i = 0; i < [pegBoard spaceCount]; i++) {
        pegStates[i] = ((i % 2) == 0) ? YES : NO;
    }
    
    recalcPegPaths = YES;
    selectedPeg = CHPegBoard_NO_PEG;
    dragPeg = CHPegBoard_NO_PEG;

    [self setNeedsDisplay: YES];

} // setPegBoard


@end // CHPegBoardView

/* $Id: CHPegBoardView.m,v 1.1 2004/05/07 22:43:36 borkware Exp $ */

